@@ -18,6 +18,7 @@
 #include "vcardtool_p.h"
 
 #include <QString>
+#include <QTimeZone>
 
 using namespace KContacts;
 
@@ -87,13 +88,9 @@ static PhoneNumber::TypeFlag stringToPhoneType(const QString &str)
     return it != std::end(s_phoneTypes) ? it->flag : PhoneNumber::TypeFlag{};
 }
 
-VCardTool::VCardTool()
-{
-}
+VCardTool::VCardTool() = default;
 
-VCardTool::~VCardTool()
-{
-}
+VCardTool::~VCardTool() = default;
 
 QByteArray VCardTool::exportVCards(const Addressee::List &list, VCard::Version version) const
 {
@@ -168,9 +165,7 @@ void VCardTool::processAddresses(const Address::List &addresses, VCard::Version
 
         if (hasLabel) {
             if (version == VCard::v4_0) {
-                if (!addr.label().isEmpty()) {
-                    adrLine.addParameter(QStringLiteral("LABEL"), QStringLiteral("\"%1\"").arg(addr.label()));
-                }
+                adrLine.addParameter(QStringLiteral("LABEL"), QStringLiteral("\"%1\"").arg(addr.label()));
             } else {
                 VCardLine labelLine(QStringLiteral("LABEL"), addr.label());
                 if (version == VCard::v2_1 && needsEncoding(addr.label())) {
@@ -317,20 +312,18 @@ void VCardTool::processCustoms(const QStringList &customs, VCard::Version versio
 
         if (identifier.toLower() == QLatin1String("x-kaddressbook-x-anniversary") && version == VCard::v4_0) {
             // ANNIVERSARY
-            if (!value.isEmpty()) {
-                const QDate date = QDate::fromString(value, Qt::ISODate);
-                QDateTime dt = QDateTime(date.startOfDay());
-                dt.setTime(QTime());
-                VCardLine line(QStringLiteral("ANNIVERSARY"), createDateTime(dt, version, false));
-                card->addLine(line);
-            }
+            const QDate date = QDate::fromString(value, Qt::ISODate);
+            QDateTime dt = QDateTime(date.startOfDay());
+            dt.setTime(QTime());
+            VCardLine line(QStringLiteral("ANNIVERSARY"));
+            line.addParameter(QStringLiteral("VALUE"), QStringLiteral("DATE"));
+            line.setValue(createDateTime(dt, version, true));
+            card->addLine(line);
         } else if (identifier.toLower() == QLatin1String("x-kaddressbook-x-spousesname") && version == VCard::v4_0) {
-            if (!value.isEmpty()) {
-                VCardLine line(QStringLiteral("RELATED"), QStringLiteral(";"));
-                line.addParameter(QStringLiteral("TYPE"), QStringLiteral("spouse"));
-                line.addParameter(QStringLiteral("VALUE"), value);
-                card->addLine(line);
-            }
+            VCardLine line(QStringLiteral("RELATED"), QStringLiteral(";"));
+            line.addParameter(QStringLiteral("TYPE"), QStringLiteral("spouse"));
+            line.addParameter(QStringLiteral("VALUE"), value);
+            card->addLine(line);
         } else {
             VCardLine line(identifier, value);
             if (version == VCard::v2_1 && needsEncoding(value)) {
@@ -362,10 +355,23 @@ QByteArray VCardTool::createVCards(const Addressee::List &list, VCard::Version v
         processAddresses(addresses, version, &card);
 
         // BDAY
-        const bool withTime = addressee.birthdayHasTime();
-        const QString birthdayString = createDateTime(addressee.birthday(), version, withTime);
-        card.addLine(VCardLine(QStringLiteral("BDAY"), birthdayString));
-
+        if (version == VCard::v4_0) {
+            VCardLine line(QStringLiteral("BDAY"));
+            line.addParameter(QStringLiteral("VALUE"), QStringLiteral("DATE"));
+            const bool withTime = addressee.birthdayHasTime();
+            line.setValue(createDateTime(addressee.birthday(), version, withTime));
+            card.addLine(line);
+        } else {
+            const bool withTime = addressee.birthdayHasTime();
+            const QString birthdayString = createDateTime(addressee.birthday(), version, withTime);
+            card.addLine(VCardLine(QStringLiteral("BDAY"), birthdayString));
+        }
+        // DEATHDATE
+        if (version == VCard::v4_0) {
+            VCardLine line(QStringLiteral("DEATHDATE"));
+            line.setValue(createDateTime(addressee.deathDate(), version));
+            card.addLine(line);
+        }
         // CATEGORIES only > 2.1
         if (version != VCard::v2_1) {
             QStringList categories = addressee.categories();
@@ -411,14 +417,14 @@ QByteArray VCardTool::createVCards(const Addressee::List &list, VCard::Version v
                 line.addParameters(lang.params());
                 card.addLine(line);
             }
-        }
-        // CLIENTPIDMAP
-        if (version == VCard::v4_0) {
-            const ClientPidMap::List clientpidmapList = addressee.clientPidMapList();
-            for (const auto &pMap : clientpidmapList) {
-                VCardLine line(QStringLiteral("CLIENTPIDMAP"), pMap.clientPidMap());
-                line.addParameters(pMap.params());
-                card.addLine(line);
+            // CLIENTPIDMAP
+            if (version == VCard::v4_0) {
+                const ClientPidMap::List clientpidmapList = addressee.clientPidMapList();
+                for (const auto &pMap : clientpidmapList) {
+                    VCardLine line(QStringLiteral("CLIENTPIDMAP"), pMap.clientPidMap());
+                    line.addParameters(pMap.params());
+                    card.addLine(line);
+                }
             }
         }
         // EMAIL
@@ -452,10 +458,18 @@ QByteArray VCardTool::createVCards(const Addressee::List &list, VCard::Version v
         }
 
         // LOGO
-        card.addLine(createPicture(QStringLiteral("LOGO"), addressee.logo(), version));
-        const QList<Picture> lstLogo = addressee.extraLogoList();
-        for (const Picture &logo : lstLogo) {
-            card.addLine(createPicture(QStringLiteral("LOGO"), logo, version));
+        if (version == VCard::v4_0) {
+            card.addLine(createPicturev4(QStringLiteral("LOGO"), addressee.logo()));
+            const QList<Picture> lstLogo = addressee.extraLogoList();
+            for (const Picture &logo : lstLogo) {
+                card.addLine(createPicturev4(QStringLiteral("LOGO"), logo));
+            }
+        } else {
+            card.addLine(createPicture(QStringLiteral("LOGO"), addressee.logo(), version));
+            const QList<Picture> lstLogo = addressee.extraLogoList();
+            for (const Picture &logo : lstLogo) {
+                card.addLine(createPicture(QStringLiteral("LOGO"), logo, version));
+            }
         }
 
         // MAILER only for version < 4.0
@@ -520,10 +534,18 @@ QByteArray VCardTool::createVCards(const Addressee::List &list, VCard::Version v
         processOrganizations(addressee, version, &card);
 
         // PHOTO
-        card.addLine(createPicture(QStringLiteral("PHOTO"), addressee.photo(), version));
-        const QList<Picture> lstExtraPhoto = addressee.extraPhotoList();
-        for (const Picture &photo : lstExtraPhoto) {
-            card.addLine(createPicture(QStringLiteral("PHOTO"), photo, version));
+        if (version == VCard::v4_0) {
+            card.addLine(createPicturev4(QStringLiteral("PHOTO"), addressee.photo()));
+            const QList<Picture> lstExtraPhoto = addressee.extraPhotoList();
+            for (const Picture &photo : lstExtraPhoto) {
+                card.addLine(createPicturev4(QStringLiteral("PHOTO"), photo));
+            }
+        } else {
+            card.addLine(createPicture(QStringLiteral("PHOTO"), addressee.photo(), version));
+            const QList<Picture> lstExtraPhoto = addressee.extraPhotoList();
+            for (const Picture &photo : lstExtraPhoto) {
+                card.addLine(createPicture(QStringLiteral("PHOTO"), photo, version));
+            }
         }
 
         // PROID only for version > 2.1
@@ -552,10 +574,14 @@ QByteArray VCardTool::createVCards(const Addressee::List &list, VCard::Version v
         }
 
         // SOUND
-        card.addLine(createSound(addressee.sound(), version));
-        const QList<Sound> lstSound = addressee.extraSoundList();
-        for (const Sound &sound : lstSound) {
-            card.addLine(createSound(sound, version));
+        if (version == VCard::v4_0) {
+            // TODO
+        } else {
+            card.addLine(createSound(addressee.sound(), version));
+            const QList<Sound> lstSound = addressee.extraSoundList();
+            for (const Sound &sound : lstSound) {
+                card.addLine(createSound(sound, version));
+            }
         }
 
         // TEL
@@ -619,9 +645,6 @@ QByteArray VCardTool::createVCards(const Addressee::List &list, VCard::Version v
                 VCardLine line(QStringLiteral("KIND"), addressee.kind());
                 card.addLine(line);
             }
-        }
-        // From vcard4.
-        if (version == VCard::v4_0) {
             const QList<CalendarUrl> lstCalendarUrl = addressee.calendarUrlList();
             for (const CalendarUrl &url : lstCalendarUrl) {
                 if (url.isValid()) {
@@ -782,6 +805,11 @@ Addressee::List VCardTool::parseVCards(const QByteArray &vcard) const
                     const QDateTime bday = parseDateTime((*lineIt).value().toString(), &withTime);
                     addr.setBirthday(bday, withTime);
                 }
+                // DEATHDATE
+                else if (identifier == QLatin1String("deathdate")) {
+                    const QDateTime deathDate = parseDateTime((*lineIt).value().toString());
+                    addr.setDeathDate(deathDate);
+                }
                 // CATEGORIES
                 else if (identifier == QLatin1String("categories")) {
                     const QStringList categories = splitString(commaSep, (*lineIt).value().toString());
@@ -854,7 +882,6 @@ Addressee::List VCardTool::parseVCards(const QByteArray &vcard) const
                 }
                 // EMAIL
                 else if (identifier == QLatin1String("email")) {
-                    const QStringList types = (*lineIt).parameters(QStringLiteral("type"));
                     Email mail((*lineIt).value().toString());
                     mail.setParams((*lineIt).parameterMap());
                     addr.addEmail(mail);
@@ -987,7 +1014,7 @@ Addressee::List VCardTool::parseVCards(const QByteArray &vcard) const
                 }
                 // PHOTO
                 else if (identifier == QLatin1String("photo")) {
-                    Picture picture = parsePicture(*lineIt);
+                    const Picture picture = parsePicture(*lineIt);
                     if (addr.photo().isEmpty()) {
                         addr.setPhoto(picture);
                     } else {
@@ -1205,8 +1232,7 @@ QDateTime VCardTool::parseDateTime(const QString &str, bool *timeValid)
     }
 
     QTime time;
-    Qt::TimeSpec spec = Qt::LocalTime;
-    int offsetSecs = 0;
+    QTimeZone tz = QTimeZone::LocalTime;
     if (posT >= 0) {
         QString timeString = str.mid(posT + 1);
         timeString.remove(QLatin1Char(':'));
@@ -1236,10 +1262,9 @@ QDateTime VCardTool::parseDateTime(const QString &str, bool *timeValid)
 
         if (tzPos >= 0) {
             if (zPos >= 0) {
-                spec = Qt::UTC;
+                tz = QTimeZone::UTC;
             } else {
-                spec = Qt::OffsetFromUTC;
-
+                int offsetSecs = 0;
                 const auto offsetString = QStringView(timeString).mid(tzPos + 1);
                 switch (offsetString.size()) {
                 case 2: // format: "hh"
@@ -1249,9 +1274,10 @@ QDateTime VCardTool::parseDateTime(const QString &str, bool *timeValid)
                     offsetSecs = offsetString.left(2).toInt() * 3600 + offsetString.mid(2, 2).toInt() * 60;
                     break;
                 }
-            }
-            if (minusPos >= 0) {
-                offsetSecs *= -1;
+                if (minusPos >= 0) {
+                    offsetSecs *= -1;
+                }
+                tz = QTimeZone::fromSecondsAheadOfUtc(offsetSecs);
             }
         }
     }
@@ -1259,7 +1285,7 @@ QDateTime VCardTool::parseDateTime(const QString &str, bool *timeValid)
         *timeValid = time.isValid();
     }
 
-    return QDateTime(date, time, spec, offsetSecs);
+    return QDateTime(date, time, tz);
 }
 
 QString VCardTool::createDateTime(const QDateTime &dateTime, VCard::Version version, bool withTime)
@@ -1281,7 +1307,7 @@ QString VCardTool::createDateTime(const QDateTime &dateTime, VCard::Version vers
         } else {
             str += QLatin1Char('-');
         }
-        QTime offsetTime = QTime(0, 0).addSecs(abs(offsetSecs));
+        const QTime offsetTime = QTime(0, 0).addSecs(abs(offsetSecs));
         if (version == VCard::v4_0) {
             str += offsetTime.toString(QStringLiteral("HHmm"));
         } else {
@@ -1322,21 +1348,54 @@ Picture VCardTool::parsePicture(const VCardLine &line) const
     Picture pic;
 
     const QStringList params = line.parameterList();
-    QString type;
-    if (params.contains(QLatin1String("type"))) {
-        type = line.parameter(QStringLiteral("type"));
-    }
-    if (params.contains(QLatin1String("encoding"))) {
-        pic.setRawData(line.value().toByteArray(), type);
-    } else if (params.contains(QLatin1String("value"))) {
-        if (line.parameter(QStringLiteral("value")).toLower() == QLatin1String("uri")) {
-            pic.setUrl(line.value().toString());
+    if (params.isEmpty()) { // Vcard4
+        if (line.value().isValid()) { // Load picture url
+            const QByteArray value = line.value().toByteArray();
+            if (const int dataPos = value.indexOf("data:"); dataPos > -1) { // Data element
+                // data:image/png;base64,
+                const int indexOfBase64 = value.indexOf(";base64,");
+                if (indexOfBase64 > -1) {
+                    const QByteArray type = value.mid(5, indexOfBase64 - 4);
+                    const QByteArray f = value.fromBase64(value.last(value.length() - 22));
+                    pic.setRawData(f, QString::fromLatin1(type));
+                }
+            } else {
+                pic.setUrl(line.value().toString());
+            }
+        }
+    } else {
+        QString type;
+        if (params.contains(QLatin1String("type"))) {
+            type = line.parameter(QStringLiteral("type"));
+        }
+        if (params.contains(QLatin1String("encoding"))) {
+            pic.setRawData(line.value().toByteArray(), type);
+        } else if (params.contains(QLatin1String("value"))) {
+            if (line.parameter(QStringLiteral("value")).toLower() == QLatin1String("uri")) {
+                pic.setUrl(line.value().toString());
+            }
         }
     }
 
     return pic;
 }
 
+VCardLine VCardTool::createPicturev4(const QString &identifier, const Picture &pic) const
+{
+    if (pic.isIntern()) {
+        const QString pictureIdentifier = identifier + QStringLiteral(":data") + QStringLiteral(":image/") + pic.type();
+        VCardLine line(pictureIdentifier);
+        line.addParameter(QStringLiteral("type"), pic.type());
+        line.setBase64Value(pic.rawData().toBase64());
+        return line;
+    } else {
+        VCardLine line(identifier);
+        line.setValue(pic.url());
+        return line;
+    }
+    return {};
+}
+
 VCardLine VCardTool::createPicture(const QString &identifier, const Picture &pic, VCard::Version version) const
 {
     VCardLine line(identifier);
@@ -1347,16 +1406,18 @@ VCardLine VCardTool::createPicture(const QString &identifier, const Picture &pic
 
     if (pic.isIntern()) {
         line.setValue(pic.rawData());
-        if (version == VCard::v2_1) {
+        switch (version) {
+        case VCard::v2_1:
             line.addParameter(QStringLiteral("ENCODING"), QStringLiteral("BASE64"));
             line.addParameter(pic.type(), QString());
-        } else { /*if (version == VCard::v3_0) */
+            break;
+        case VCard::v3_0:
             line.addParameter(QStringLiteral("encoding"), QStringLiteral("b"));
             line.addParameter(QStringLiteral("type"), pic.type());
-#if 0
-        } else { //version 4.0
-            line.addParameter(QStringLiteral("data") + QStringLiteral(":image/") + pic.type(), QStringLiteral("base64"));
-#endif
+            break;
+        case VCard::v4_0:
+            // Nothing use createPicturev4
+            break;
         }
     } else {
         line.setValue(pic.url());
@@ -1389,7 +1450,6 @@ Sound VCardTool::parseSound(const VCardLine &line) const
 
 VCardLine VCardTool::createSound(const Sound &snd, VCard::Version version) const
 {
-    Q_UNUSED(version);
     VCardLine line(QStringLiteral("SOUND"));
 
     if (snd.isIntern()) {
